#ifdef _WIN32
#define NOMINMAX
#endif

#include "Imagine/Graphics/ConfigGraphics.h"
#include "Imagine/Graphics/Mesh.h"
#include <numeric>
#include <sstream>

#include <cmath>

#ifdef IMAGINE_OPENGL

// No use of QOpenGLWidget here, but does the platform-specific #include <gl.h>
#include <QOpenGLWidget>
#include <QOpenGLFunctions>
#include <QOpenGLShaderProgram>

namespace Imagine {

class MeshInternal : protected QOpenGLFunctions {
private:
    FloatPoint3 computeTriangleNormal(int index) const;
    FloatPoint3 computeQuadNormal(int index) const;
    FloatPoint3 computeVertexNormal(int v,
                                    const std::vector<std::vector<int> >&,
                                    const std::vector<std::vector<int> >&)const;
    void computeFaceRings(std::vector<std::vector<int> >& tRings,
                          std::vector<std::vector<int> >& qRings) const;
public:
    void computeNormals();
    void setColorsFromValues();
    void setDataShaders();

public:
    static MeshInternal* Arrow(const DoublePoint3 &base,
                               const DoubleVector3 &shaft,
                               double r, int resolution);
    static MeshInternal* Cone(const DoublePoint3 &base,
                              const DoubleVector3 &axis,
                              double r, int resolution);
    static MeshInternal* Cube(const DoublePoint3& center,
                              const DoubleVector3& axis1,
                              const DoubleVector3& axis2,
                              const DoubleVector3& axis3);
    static MeshInternal* Cylinder(const DoublePoint3 &center,
                                  const DoubleVector3 &axis,
                                  double r, int resolution);
    static MeshInternal* Plane(const DoublePoint3& center,
                               const DoubleVector3& axis1,
                               const DoubleVector3& axis2);
    static MeshInternal* PointCloud(const DoublePoint3 *points, int n);
    static MeshInternal* PolyLine(DoublePoint3 *points, int n, bool closed);
    static MeshInternal* Sphere(const DoublePoint3& center,
                                double r, int resolution);

public:
    MeshInternal();
    MeshInternal(const FloatPoint3* vertices_, int nvertices,
                 const Triangle* triangles_, int ntriangles,
                 const Quad* quads_, int nquads,
                 MeshColorMode colorMode_,
                 MeshShadingMode shadingMode_,
                 const FloatVector3* _normals);
    MeshInternal(const DoublePoint3* vertices_, int nvertices,
                 const Triangle* triangles_, int ntriangles,
                 const Quad* quads_, int nquads,
                 MeshColorMode colorMode_,
                 MeshShadingMode shadingMode_,
                 const DoubleVector3* _normals);
    MeshInternal(const std::vector<FloatPoint3>& vertices_,
                 const std::vector<Triangle>& triangles_,
                 const std::vector<Quad>& quads_,
                 MeshColorMode colorMode_,
                 MeshShadingMode shadingMode_,
                 const std::vector<FloatVector3>& normals_);

    inline const FloatPoint3& coords(int index) const
    { return vertices[index]; }
    inline const FloatVector3& normal(int index) const
    { return normals[index]; }
    inline const Triangle& triangle(int index) const
    { return triangles[index]; }
    inline const Quad& quad(int index) const
    { return quads[index]; }
    inline int triangleVertex(int vInd, int tIndex) const
    { return triangle(tIndex)[vInd]; }
    // One quad = two triangles = 6 vertices
    inline int quadVertex(int vInd, int qIndex) const
    { return quad(qIndex)[(vInd<4)? vInd: 2*vInd-8]; }
    inline const FloatPoint3& triangleCoords(int vInd, int tInd) const
    { return vertices[ triangleVertex(vInd, tInd) ]; }
    inline const FloatPoint3& quadCoords(int vInd, int qInd) const
    { return vertices[ quadVertex(vInd, qInd) ]; }

    void setGeometry(const FloatPoint3* vertices_, int nvertices,
                     const Triangle* triangles_, int ntriangles,
                     const Quad* quads_, int nquads,
                     const FloatVector3 *normals_);
    void setGeometry(const DoublePoint3* vertices_, int nvertices,
                     const Triangle* triangles_, int ntriangles,
                     const Quad* quads_, int nquads,
                     const DoubleVector3 *normals_);

    void setColor(Color col);
    void setColors(MeshData data, const Color *col);
    void setDefaultLookupTable();
    void setLookupTable(Color col0, Color col1);
    void setRange(float v0, float v1);
    void setValues(MeshData data, const float *val);

    void setOpacity(float opacity_);

    void drawMesh(MeshFaceMode faceMode, QOpenGLShaderProgram* prog,
                  const FVector<float,4>& light);

    friend bool readMeshFromObjFile(Mesh& mesh, const std::string& fileName);
public: /* data members */
    std::vector<FloatPoint3> vertices;
    std::vector<FloatPoint3> verticesT; // Vertices for triangles
    std::vector<FloatPoint3> verticesQ; // Vertices for quads
    std::vector<FloatPoint3> verticesTL; // Vertices for triangle outlines
    std::vector<FloatPoint3> verticesQL; // Vertices for quad outlines
    std::vector<FloatVector3> normals, normalsT, normalsQ;
    std::vector<Triangle> triangles;
    std::vector<Quad> quads;

    // Color of each vertex in the same order as defined in "vertices"
    // Only in CONSTANT_COLOR, VERTEX_COLOR or VERTEX_VALUE colorModes
    std::vector<Color> colors;
    std::vector<FloatPoint3> colorsP; // points
    std::vector<FloatPoint3> colorsT; // triangles
    std::vector<FloatPoint3> colorsQ; // quads
    std::vector<FloatPoint3> colorsTL; // triangles
    std::vector<FloatPoint3> colorsQL; // quads

    // Color of each triangle in the same order as defined in "triangles"
    // Only in FACE_COLOR or FACE_VALUE colorModes
    std::vector<Color> triangleColors;

    // Color of each quad in the same order as defined in "quad"
    // Only in FACE_COLOR or FACE_VALUE colorModes
    std::vector<Color> quadColors;

    // Vertex values
    // Value of each vertex or face that determines its color in VERTEX_VALUE
    std::vector<float> vertexValues;

    // Face values
    // Value of each triangle (or quad) that determines its color in FACE_VALUE
    std::vector<float> triangleValues;
    std::vector<float> quadValues;

public: /* parameters */
    // Drawing modes
    MeshColorMode colorMode;
    MeshShadingMode shadingMode;
    bool polyline, colorsSet;
    float opacity;
    float minLookupRange;
    float maxLookupRange;
    Color minLookupColor;
    Color maxLookupColor;
};

// Prepare all data for injecting as input to shaders
void MeshInternal::setDataShaders() {
    FloatPoint3 col;
    col[0] = float(colors[0].r()) / 255.0f;
    col[1] = float(colors[0].g()) / 255.0f;
    col[2] = float(colors[0].b()) / 255.0f;

    colorsP.clear();
    if(colors.size()>=vertices.size() &&
       colorsSet && (colorMode==VERTEX_COLOR || colorMode==VERTEX_VALUE))
        for(size_t i=0; i<vertices.size(); i++) {
            col[0] = float(colors[i].r()) / 255.0f;
            col[1] = float(colors[i].g()) / 255.0f;
            col[2] = float(colors[i].b()) / 255.0f;
            colorsP.push_back(col);
        }
    else
        colorsP.insert(colorsP.begin(), vertices.size(), col);

    FloatVector3 n;
    verticesT.clear(); colorsT.clear(); normalsT.clear();
    for(size_t t=0; t<triangles.size(); ++t) {
        if(!triangleColors.empty() &&
           colorsSet && (colorMode==FACE_COLOR || colorMode==FACE_VALUE)) {
            col[0] = float(triangleColors[t].r()) / 255.0f;
            col[1] = float(triangleColors[t].g()) / 255.0f;
            col[2] = float(triangleColors[t].b()) / 255.0f;
        }
        if(shadingMode == FLAT_SHADING)
            n = computeTriangleNormal((int)t);
        for(int v=0; v<3; ++v) {
            int vInd = triangleVertex(v,(int)t);
            if(colorsSet && (colorMode==VERTEX_COLOR ||
                             colorMode==VERTEX_VALUE)) {
                col[0] = float(colors[vInd].r()) / 255.0f;
                col[1] = float(colors[vInd].g()) / 255.0f;
                col[2] = float(colors[vInd].b()) / 255.0f;
            }
            if(shadingMode == SMOOTH_SHADING)
                n = normal(vInd);
            normalsT.push_back(n);
            colorsT.push_back(col);
            verticesT.push_back(coords(vInd));
        }
    }

    verticesTL.clear(); colorsTL.clear();
    for(size_t t=0; t<triangles.size(); ++t) {
        if(!triangleColors.empty() &&
           colorsSet && (colorMode==FACE_COLOR || colorMode==FACE_VALUE)) {
            col[0] = float(triangleColors[t].r()) / 255.0f;
            col[1] = float(triangleColors[t].g()) / 255.0f;
            col[2] = float(triangleColors[t].b()) / 255.0f;
        }
        for(int v=0; v<6; ++v) {
            int vInd = triangleVertex(((v+1)>>1)%3,(int)t);
            if(colorsSet && (colorMode==VERTEX_COLOR ||
                             colorMode==VERTEX_VALUE)) {
                col[0] = float(colors[vInd].r()) / 255.0f;
                col[1] = float(colors[vInd].g()) / 255.0f;
                col[2] = float(colors[vInd].b()) / 255.0f;
            }
            colorsTL.push_back(col);
            verticesTL.push_back(coords(vInd));
        }
    }

    verticesQ.clear(); colorsQ.clear(); normalsQ.clear();
    for(size_t q=0; q<quads.size(); ++q) {
        if(!quadColors.empty() &&
           colorsSet && (colorMode==FACE_COLOR || colorMode==FACE_VALUE)) {
            col[0] = float(quadColors[q].r()) / 255.0f;
            col[1] = float(quadColors[q].g()) / 255.0f;
            col[2] = float(quadColors[q].b()) / 255.0f;
        }
        if(shadingMode == FLAT_SHADING)
            n = computeQuadNormal((int)q);
        for(int v=0; v<6; ++v) {
            int vInd = quadVertex(v,(int)q);
            if(colorsSet && (colorMode==VERTEX_COLOR ||
                             colorMode==VERTEX_VALUE)) {
                col[0] = float(colors[vInd].r()) / 255.0f;
                col[1] = float(colors[vInd].g()) / 255.0f;
                col[2] = float(colors[vInd].b()) / 255.0f;
            }
            if(shadingMode == SMOOTH_SHADING)
                n = normal(vInd);
            normalsQ.push_back(n);
            colorsQ.push_back(col);
            verticesQ.push_back(coords(vInd).data());
        }
    }

    verticesQL.clear(); colorsQL.clear();
    for(size_t t=0; t<quads.size(); ++t) {
        if(!quadColors.empty() &&
           colorsSet && (colorMode==FACE_COLOR || colorMode==FACE_VALUE)) {
            col[0] = float(quadColors[t].r()) / 255.0f;
            col[1] = float(quadColors[t].g()) / 255.0f;
            col[2] = float(quadColors[t].b()) / 255.0f;
        }
        for(int v=0; v<8; ++v) {
            int vInd = quadVertex(((v+1)>>1)%4,(int)t);
            if(colorsSet && (colorMode==VERTEX_COLOR ||
                             colorMode==VERTEX_VALUE)) {
                col[0] = float(colors[vInd].r()) / 255.0f;
                col[1] = float(colors[vInd].g()) / 255.0f;
                col[2] = float(colors[vInd].b()) / 255.0f;
            }
            colorsQL.push_back(col);
            verticesQL.push_back(coords(vInd));
        }
    }
}

FloatPoint3 MeshInternal::computeTriangleNormal(int index) const {
    FloatPoint3 u(triangleCoords(1,index));
    u -= triangleCoords(0,index);
    FloatPoint3 v(triangleCoords(2,index));
    v -= triangleCoords(0,index);
    FloatPoint3 n = u^v;
    n.normalize();
    return n;
}

FloatPoint3 MeshInternal::computeQuadNormal(int index) const {
    FloatPoint3 u(quadCoords(1,index));
    u -= quadCoords(0,index);
    FloatPoint3 v(quadCoords(2,index));
    v -= quadCoords(0,index);
    FloatPoint3 n = u^v;
    n.normalize();
    return n;
}

FloatPoint3 MeshInternal::computeVertexNormal(int v,
                        const std::vector<std::vector<int> >& triRings,
                        const std::vector<std::vector<int> >& quadRings) const {
    FloatVector3 n(0.f);
    const std::vector<int>& faceTriangleRing = triRings[v];
    const std::vector<int>& faceQuadRing = quadRings[v];
    for(size_t t=0; t<faceTriangleRing.size(); ++t)
        n += computeTriangleNormal(faceTriangleRing[t]);
    for(size_t t=0; t<faceQuadRing.size(); ++t)
        n += computeQuadNormal(faceQuadRing[t]);
    n /= static_cast<float>(faceTriangleRing.size() + faceQuadRing.size());
    n.normalize();
    return n;
}

void MeshInternal::computeFaceRings(std::vector<std::vector<int> >& tRings,
                                    std::vector<std::vector<int> >& qRings)const
{
    tRings.resize(vertices.size());
    qRings.resize(vertices.size());

    for (int t=0; t<int(triangles.size()); ++t) {
        tRings[triangleVertex(0,t)].push_back(t);
        tRings[triangleVertex(1,t)].push_back(t);
        tRings[triangleVertex(2,t)].push_back(t);
    }
    for (int q=0; q<int(quads.size()); ++q) {
        qRings[quadVertex(0,q)].push_back(q);
        qRings[quadVertex(1,q)].push_back(q);
        qRings[quadVertex(2,q)].push_back(q);
        qRings[quadVertex(3,q)].push_back(q);
    }
}

void MeshInternal::computeNormals() {
    if(vertices.empty() || (triangles.empty() && quads.empty()))
        return;
    if(shadingMode == FLAT_SHADING)
        return; // Normals are not on vertices but on faces

    // For each vertex, adjacent/immediate neighboring triangles/quads
    std::vector<std::vector<int> > tRings;
    std::vector<std::vector<int> > qRings;

    computeFaceRings(tRings, qRings);
    normals.resize(vertices.size());
    for(int v=0; v<int(vertices.size()); ++v)
        normals[v] = computeVertexNormal(v, tRings, qRings);
}

MeshInternal::MeshInternal() {
    colorMode = CONSTANT_COLOR;
    shadingMode = SMOOTH_SHADING;
    polyline = colorsSet = false;
    colors.push_back(Color(127,127,127));
    opacity = 1.0f;
}

MeshInternal::MeshInternal(const FloatPoint3* vertices_, int nvertices,
                           const Triangle* triangles_, int ntriangles,
                           const Quad* quads_, int nquads,
                           MeshColorMode colorMode_,
                           MeshShadingMode shadingMode_,
                           const FloatVector3* _normals)
: colorMode(colorMode_), shadingMode(shadingMode_), polyline(false)
, colorsSet(false), opacity(1.0f)
{
    colors.push_back(Color(127,127,127));
    setDefaultLookupTable();
    setGeometry(vertices_, nvertices, triangles_, ntriangles,
                quads_, nquads, _normals);
}

MeshInternal::MeshInternal(const DoublePoint3* vertices_, int nvertices,
                           const Triangle *triangles_, int ntriangles,
                           const Quad *quads_, int nquads,
                           MeshColorMode colorMode_,
                           MeshShadingMode shadingMode_,
                           const DoubleVector3* _normals)
: colorMode(colorMode_), shadingMode(shadingMode_), polyline(false)
, colorsSet(false), opacity(1.0f)
{
    colors.push_back(Color(127,127,127));
    setDefaultLookupTable();
    setGeometry(vertices_, nvertices, triangles_, ntriangles,
                quads_, nquads, _normals);
}

MeshInternal::MeshInternal(const std::vector<FloatPoint3>& vertices_,
                           const std::vector<Triangle>& triangles_,
                           const std::vector<Quad>& quads_,
                           MeshColorMode colorMode_,
                           MeshShadingMode shadingMode_,
                           const std::vector<FloatVector3>& normals_)
: vertices(vertices_),normals(normals_),triangles(triangles_),quads(quads_)
, colorMode(colorMode_), shadingMode(shadingMode_), polyline(false)
, colorsSet(false), opacity(1.0f)
{
    colors.push_back(Color(127,127,127));
    setDefaultLookupTable();
    setDataShaders();
}

void MeshInternal::setGeometry(const FloatPoint3* vertices_, int nvertices,
                               const Triangle* triangles_, int ntriangles,
                               const Quad* quads_, int nquads,
                               const FloatVector3 *normals_) {
    vertices.clear();
    for(int i=0; i<nvertices; i++)
        vertices.push_back(vertices_[i]);
    triangles.clear();
    for(int i=0; i<ntriangles; i++)
        triangles.push_back(triangles_[i]);
    quads.clear();
    for(int i=0; i<nquads; i++)
        quads.push_back(quads_[i]);
    if(normals_)
        for(int i=0; i<nvertices; i++)
            normals.push_back(normals_[i]);
    else
        computeNormals();
    setDataShaders();
}

void MeshInternal::setGeometry(const DoublePoint3* vertices_, int nvertices,
                               const Triangle* triangles_, int ntriangles,
                               const Quad* quads_, int nquads,
                               const DoubleVector3* normals_) {
    vertices.clear();
    for(int i=0; i<nvertices; i++)
        vertices.push_back(vertices_[i]);
    triangles.clear();
    for(int i=0; i<ntriangles; i++)
        triangles.push_back(triangles_[i]);
    quads.clear();
    for(int i=0; i<nquads; i++)
        quads.push_back(quads_[i]);
    if(normals_)
        for(int i=0; i<nvertices; i++)
            normals.push_back(normals_[i]);
    else
        computeNormals();
    setDataShaders();
}

void Mesh::setGeometry(const FloatPoint3* vertices_, int nvertices,
                       const Triangle* triangles_, int ntriangles,
                       const Quad* quads_, int nquads,
                       const FloatVector3 *normals_) {
    priv->setGeometry(vertices_,nvertices,triangles_,ntriangles,quads_,nquads,
                      normals_);
}
void Mesh::setGeometry(const DoublePoint3* vertices_, int nvertices,
                       const Triangle* triangles_, int ntriangles,
                       const Quad* quads_, int nquads,
                       const DoubleVector3 *normals_) {
    priv->setGeometry(vertices_,nvertices,triangles_,ntriangles,quads_,nquads,
                      normals_);
}

void Mesh::setVertices(const FloatPoint3 *vertices_,
                       const FloatVector3 *normals_) {
    size_t nvertices = priv->vertices.size();
    priv->vertices.clear();
    for(size_t i=0; i<nvertices; i++)
        priv->vertices.push_back(vertices_[i]);
    priv->normals.clear();
    if(normals_)
        for(size_t i=0; i<nvertices; i++)
            priv->normals.push_back(normals_[i]);
    else
        priv->computeNormals();
    priv->setDataShaders();
}
void Mesh::setVertices(const DoublePoint3 *vertices_,
                       const DoubleVector3 *normals_) {
    size_t nvertices = priv->vertices.size();
    priv->vertices.clear();
    for(size_t i=0; i<nvertices; i++)
        priv->vertices.push_back(vertices_[i]);
    priv->normals.clear();
    if(normals_)
        for(size_t i=0; i<nvertices; i++)
            priv->normals.push_back(normals_[i]);
    else
        priv->computeNormals();
    priv->setDataShaders();
}
const std::vector<FloatPoint3>& Mesh::vertices() const {
    return priv->vertices;
}

void Mesh::setColor(Color col) {
    if(priv->colorMode==CONSTANT_COLOR) {
        priv->colors.clear();
        priv->colors.push_back(col);
        priv->setDataShaders();
    }
}

void Mesh::setColors(MeshData data, const Color *col) {
    if(priv->colorMode==CONSTANT_COLOR ||
       priv->colorMode==VERTEX_VALUE ||
       priv->colorMode==FACE_VALUE)
        return;
    if(priv->colorMode==VERTEX_COLOR && data!=VERTEX) {
        std::cerr << "Mesh::setColors: incorrect assignment" << std::endl;
        return;
    }
    if(priv->colorMode==FACE_COLOR && data==VERTEX) {
        std::cerr << "Mesh::setColors: incorrect assignment" << std::endl;
        return;
    }

    switch(data) {
    case VERTEX:
        if(priv->colorMode==VERTEX_COLOR) {
            priv->colorsSet = true;
            priv->colors.clear();
            for(size_t i=0; i<priv->vertices.size(); i++)
                priv->colors.push_back(col[i]);
            priv->setDataShaders();
        }
        break;
    case TRIANGLE:
        if(priv->colorMode==FACE_COLOR) {
            priv->colorsSet = true;
            priv->triangleColors.clear();
            for(size_t i=0; i<priv->triangles.size(); i++)
                priv->triangleColors.push_back(col[i]);
            priv->setDataShaders();
        }
        break;
    case QUAD:
        if(priv->colorMode==FACE_COLOR) {
            priv->colorsSet = true;
            priv->quadColors.clear();
            for(size_t i=0; i<priv->quads.size(); i++)
                priv->quadColors.push_back(col[i]);
            priv->setDataShaders();
        }
        break;
    }
}

void MeshInternal::setDefaultLookupTable() {
    if(colorMode==VERTEX_VALUE || colorMode==FACE_VALUE) {
        minLookupColor = Color(85,85,255);
        maxLookupColor = Color(255, 0, 0);
        minLookupRange = 0.0f;
        maxLookupRange = 1.0f;
        setColorsFromValues();
    }
}
void Mesh::setDefaultLookupTable() {
    priv->setDefaultLookupTable();
    priv->setDataShaders();
}

void Mesh::setLookupTable(Color col0, Color col1) {
    if(priv->colorMode==VERTEX_VALUE || priv->colorMode==FACE_VALUE) {
        priv->minLookupColor = col0;
        priv->maxLookupColor = col1;
        priv->setColorsFromValues();
        priv->setDataShaders();
    }
}

void Mesh::setRange(float v0, float v1) {
    if(priv->colorMode==VERTEX_VALUE || priv->colorMode==FACE_VALUE) {
        if(v0 >= v1)
            return;
        priv->minLookupRange = v0;
        priv->maxLookupRange = v1;
    }
}

void MeshInternal::setColorsFromValues() {
    unsigned minR = std::min(minLookupColor.r(),maxLookupColor.r());
    unsigned maxR = std::max(minLookupColor.r(),maxLookupColor.r());
    unsigned minG = std::min(minLookupColor.g(),maxLookupColor.g());
    unsigned maxG = std::max(minLookupColor.g(),maxLookupColor.g());
    unsigned minB = std::min(minLookupColor.b(),maxLookupColor.b());
    unsigned maxB = std::max(minLookupColor.b(),maxLookupColor.b());
    unsigned int R,G,B;
    const float f = 1.0f/(maxLookupRange-minLookupRange);
    if(! vertexValues.empty())
        colors.clear();
    for(size_t i=0; i<vertexValues.size(); i++) {
        float v = vertexValues[i];
        R = (v*f)*maxR + (1-v*f)*minR;
        G = (v*f)*maxG + (1-v*f)*minG;
        B = (v*f)*maxB + (1-v*f)*minB;
        colors.push_back(Color(R, G, B));
    }
    triangleColors.clear();
    for(size_t i=0; i<triangleValues.size(); i++) {
        float v = triangleValues[i];
        R = (v*f)*maxR + (1-v*f)*minR;
        G = (v*f)*maxG + (1-v*f)*minG;
        B = (v*f)*maxB + (1-v*f)*minB;
        triangleColors.push_back(Color(R, G, B));
    }
    quadColors.clear();
    for(size_t i=0; i<quadValues.size(); i++) {
        float v = quadValues[i];
        R = (v*f)*maxR + (1-v*f)*minR;
        G = (v*f)*maxG + (1-v*f)*minG;
        B = (v*f)*maxB + (1-v*f)*minB;
        quadColors.push_back(Color(R, G, B));
    }
}

void Mesh::setValues(MeshData data, const float *val) {
    if(! (priv->colorMode==VERTEX_VALUE || priv->colorMode==FACE_VALUE))
        return;
    switch(data) {
    case VERTEX:
        priv->colorsSet = true;
        priv->vertexValues.clear();
        for(size_t i=0; i<priv->vertices.size(); i++)
            priv->vertexValues.push_back(val[i]);
        break;
    case TRIANGLE:
        priv->colorsSet = true;
        priv->triangleValues.clear();
        for(size_t i=0; i<priv->triangles.size(); i++)
            priv->triangleValues.push_back(val[i]);
        break;
    case QUAD:
        priv->colorsSet = true;
        priv->quadValues.clear();
        for(size_t i=0; i<priv->quads.size(); i++)
            priv->quadValues.push_back(val[i]);
        break;
    }
    priv->setColorsFromValues();
    priv->setDataShaders();
}

void Mesh::setOpacity(float opacity_) {
    if(0<=opacity_ && opacity_<=1)
        priv->opacity = opacity_;
    else priv->opacity = 1.0f;
}
float Mesh::getOpacity() const { return priv->opacity; }

// static meshes given for convenience.

MeshInternal* MeshInternal::Arrow(const DoublePoint3 &base,
                                 const DoubleVector3 &shaft,
                                 double r, int resolution) {
    if(norm(shaft) - 2*r <= 0) {
        std::cerr << "Incompatible value of r: please make sure norm(shaft) - 2*r > 0" << std::endl;
        return new MeshInternal;
    }
    MeshInternal* cone = Cone(base + (1-((2*r)/norm(shaft)))*shaft,
                              ((2*r)/norm(shaft))*shaft,
                              r, resolution);
    MeshInternal* arrow = Cylinder(base + ((1 - ((2*r)/norm(shaft)))/2)*shaft,
                                   (1 - ((2*r)/norm(shaft)))*shaft,
                                   r/2, resolution);

    int n = static_cast<int>(arrow->vertices.size());
    for(size_t i=0; i<cone->vertices.size(); i++)
        arrow->vertices.push_back(cone->vertices[i]);
    for(size_t i=0; i<cone->triangles.size(); i++)
        arrow->triangles.push_back(Triangle(n+cone->triangleVertex(0,(int)i),
                                            n+cone->triangleVertex(1,(int)i),
                                            n+cone->triangleVertex(2,(int)i)));
    delete cone;
    arrow->computeNormals();
    arrow->setDataShaders();
    return arrow;
}

MeshInternal* MeshInternal::Cone(const DoublePoint3 &base,
                                 const DoubleVector3 &axis,
                                 double r, int resolution) {
    MeshInternal* cone = new MeshInternal;
    FloatPoint3 point;
    double Pi = 3.14159;
    // parameters for rotation
    double axisNorm = std::sqrt(axis.x()*axis.x() + axis.y()*axis.y()+ axis.z()*axis.z());
    double xyAxisNorm = std::sqrt(axis.x()*axis.x() + axis.y()*axis.y());
    double cosTheta = axis.z() / axisNorm;
    double sinTheta = xyAxisNorm / axisNorm;
    double cosPhi = axis.x() / xyAxisNorm;
    double sinPhi = axis.y() / xyAxisNorm;
    double RyArray[9] = {cosTheta, 0., sinTheta, 0., 1., 0., -sinTheta, 0., cosTheta};
    FMatrix<double,3,3> Ry(RyArray);
    double RzArray[9] = {cosPhi, sinPhi, 0., -sinPhi, cosPhi, 0., 0., 0., 1.};
    FMatrix<double,3,3> Rz(RzArray);

    // Setting  side vertices
    for(int i=0; i<resolution; i++) {
        point.z() = 0.;
        point.x() = r*cos((2*i*Pi)/resolution);
        point.y() = r*sin((2*i*Pi)/resolution);
        // Rotation to fit to given axis
        if(xyAxisNorm != 0)
            point = Rz*(Ry*point);
        // Push back
        cone->vertices.push_back(point);
    }
    // Setting top vertex
    cone->vertices.push_back(axis);
    // Setting base vertex
    cone->vertices.push_back(FloatPoint3(0., 0., 0.));

    // Translate to fit base
    for(size_t i=0; i<cone->vertices.size(); i++)
        cone->vertices[i] += base;

    // Setting side triangles
    for(int i=0; i<resolution; i++)
        cone->triangles.push_back(Triangle((i+1)%resolution, i, resolution));

    // Setting base triangles
    for(int i=0; i<resolution; i++)
        cone->triangles.push_back(Triangle(i, (i+1)%resolution, resolution+1));

    cone->computeNormals();
    cone->setDataShaders();

    return cone;
}

MeshInternal* MeshInternal::Cube(const DoublePoint3& center,
                                 const DoubleVector3& axis1,
                                 const DoubleVector3& axis2,
                                 const DoubleVector3& axis3) {
    MeshInternal* cube = new MeshInternal;
    cube->shadingMode = FLAT_SHADING;
    FloatPoint3 point = center - (axis1 / 2) - (axis2 / 2) - (axis3 / 2);

    // Setting vertices
    cube->vertices.push_back(point);
    point += axis1;
    cube->vertices.push_back(point);
    point += axis2;
    cube->vertices.push_back(point);
    point -= axis1;
    cube->vertices.push_back(point);
    point -= axis2;
    point += axis3;
    cube->vertices.push_back(point);
    point += axis1;
    cube->vertices.push_back(point);
    point += axis2;
    cube->vertices.push_back(point);
    point -= axis1;
    cube->vertices.push_back(point);

    // Setting quads
    cube->quads.push_back(Quad(3,2,1,0));
    cube->quads.push_back(Quad(4,5,6,7));
    cube->quads.push_back(Quad(0,1,5,4));
    cube->quads.push_back(Quad(1,2,6,5));
    cube->quads.push_back(Quad(2,3,7,6));
    cube->quads.push_back(Quad(3,0,4,7));

    cube->computeNormals();
    cube->setDataShaders();
    return cube;
}

MeshInternal* MeshInternal::Cylinder(const DoublePoint3 &center,
                                     const DoubleVector3 &axis,
                                     double r, int resolution) {
    MeshInternal* cylinder = new MeshInternal;
    FloatPoint3 point;
    double Pi = 3.14159;
    // parameters for rotation
    double axisNorm = std::sqrt(axis.x()*axis.x() + axis.y()*axis.y()+ axis.z()*axis.z());
    double xyAxisNorm = std::sqrt(axis.x()*axis.x() + axis.y()*axis.y());
    double cosTheta = axis.z() / axisNorm;
    double sinTheta = xyAxisNorm / axisNorm;
    double cosPhi = axis.x() / xyAxisNorm;
    double sinPhi = axis.y() / xyAxisNorm;
    double RyArray[9] = {cosTheta, 0., sinTheta, 0., 1., 0., -sinTheta, 0., cosTheta};
    FMatrix<double,3,3> Ry(RyArray);
    double RzArray[9] = {cosPhi, sinPhi, 0., -sinPhi, cosPhi, 0., 0., 0., 1.};
    FMatrix<double,3,3> Rz(RzArray);

    // Setting bottom vertices
    for(int i=0; i<resolution; i++) {
        point.z() = -axisNorm/2;
        point.x() = r*cos((2*i*Pi)/resolution);
        point.y() = r*sin((2*i*Pi)/resolution);
        // Rotation to fit to given axis
        if(xyAxisNorm != 0)
            point = Rz*(Ry*point);
        // Push back
        cylinder->vertices.push_back(point);
    }
    // Setting top vertices
    for(int i=0; i<resolution; i++) {
        point.z() = axisNorm/2;
        point.x() = r*cos((2*i*Pi)/resolution);
        point.y() = r*sin((2*i*Pi)/resolution);
        // Rotation to fit to given axis
        if(xyAxisNorm != 0)
            point = Rz*(Ry*point);
        // Push back
        cylinder->vertices.push_back(point);
    }
    // Setting center vertices
    // Bottom
    point = -FloatPoint3(0, 0, axisNorm/2);
    if(xyAxisNorm != 0)
        point = Rz*(Ry*point);
    cylinder->vertices.push_back(point);
    // Top
    point = FloatPoint3(0, 0, axisNorm/2);
    if(xyAxisNorm != 0)
        point = Rz*(Ry*point);
    cylinder->vertices.push_back(point);

    // Translate to fit center
    for(size_t i=0; i<cylinder->vertices.size(); i++)
        cylinder->vertices[i] += center;

    // Setting side triangles
    for(int i=0; i<resolution; i++)
        cylinder->quads.push_back(Quad(i, (i+1)%resolution, resolution+((i+1)%resolution), resolution+i));

    // Setting bottom/top triangles
    for(int i=0; i<resolution; i++) {
        cylinder->triangles.push_back(Triangle(2*resolution,
                                               (i+1)%resolution, i));
        cylinder->triangles.push_back(Triangle(resolution+i,
                                               resolution+((i+1)%resolution),
                                               2*resolution+1));
    }

    cylinder->computeNormals();
    cylinder->setDataShaders();
    return cylinder;
}

MeshInternal* MeshInternal::Plane(const DoublePoint3& center,
                                  const DoubleVector3& axis1,
                                  const DoubleVector3& axis2) {
    FloatPoint3 point;
    MeshInternal* plane = new MeshInternal;

    // Setting vertices
    point = center - ((axis1 + axis2) / 2);
    plane->vertices.push_back(point);
    point += axis1;
    plane->vertices.push_back(point);
    point += axis2;
    plane->vertices.push_back(point);
    point -= axis1;
    plane->vertices.push_back(point);
    point -= axis2;
    point -= (axis1^axis2)/1000;
    plane->vertices.push_back(point);
    point += axis1;
    plane->vertices.push_back(point);
    point += axis2;
    plane->vertices.push_back(point);
    point -= axis1;
    plane->vertices.push_back(point);
    point -= axis2;

    // Setting triangles
    plane->quads.push_back(Quad(0, 1, 2, 3));
    plane->quads.push_back(Quad(7, 6, 5, 4));

    FloatVector3 normal = (axis1^axis2).normalize();
    for(int i=0; i<4; i++)
        plane->normals.push_back(normal);
    for(int i=0; i<4; i++)
        plane->normals.push_back(-normal);

    plane->setDataShaders();
    return plane;
}

MeshInternal* MeshInternal::PointCloud(const DoublePoint3 *points, int n) {
    return new MeshInternal(points,n,0,0,0,0,CONSTANT_COLOR,SMOOTH_SHADING,0);
}

// Polyline generated as a set of flat triangles.
MeshInternal* MeshInternal::PolyLine(DoublePoint3 *points, int n, bool closed) {
    MeshInternal* polyLine =
        new MeshInternal(points,n,0,0,0,0, CONSTANT_COLOR, SMOOTH_SHADING,0);
    polyLine->polyline = true;
    if(closed && n>0) {
        polyLine->vertices.push_back(points[0]);
        ++n;
    }
    polyLine->setDataShaders();
    return polyLine;
}

MeshInternal* MeshInternal::Sphere(const DoublePoint3& center,
                                   double r, int resolution) {
    MeshInternal* sphere = new MeshInternal;
    FloatPoint3 point;
    double Pi = 3.14159;

    // Setting vertices
    for(int i=1; i<resolution; i++) {
        point.z() = center.z() + r*cos(i*Pi/resolution);
        for(int j=0; j<2*resolution; j++) {
            point.x() = center.x()+ r*sin(i*Pi/resolution)*cos(j*Pi/resolution);
            point.y() = center.y()+ r*sin(i*Pi/resolution)*sin(j*Pi/resolution);
            sphere->vertices.push_back(point);
        }
    }
    // Setting bottom/top vertex
    point = center;
    point.z() += r;
    sphere->vertices.push_back(point);
    point.z() -= 2*r;
    sphere->vertices.push_back(point);

    // Setting side quads
    for(int i=0; i<resolution-2; i++) {
        for(int j=0; j<2*resolution; j++)
            sphere->quads.push_back(Quad( 2*(i+1)*resolution + j,
                                          2*(i+1)*resolution +
                                          ((j+1)%(2*resolution)),
                                          2*i*resolution +
                                          ((j+1)%(2*resolution)),
                                          2*i*resolution + j ));
    }

    // Setting bottom/top triangles
    for(int i=0; i<2*resolution; i++) {
        sphere->triangles.push_back(Triangle( i,
                                              (i+1)%(2*resolution),
                                              2*resolution*(resolution-1)));
        sphere->triangles.push_back(Triangle( 2*resolution*(resolution-1)+1,
                                              2*resolution*(resolution-2)+
                                              ((i+1)%(2*resolution)),
                                              2*resolution*(resolution-2)+i));
    }

    // Setting normals
    for(int i=0; i<2*resolution*(resolution-1)+2; i++)
        sphere->normals.push_back((sphere->vertices[i]-center).normalize());
    sphere->setDataShaders();
    return sphere;
}

// ======================================================================
// Main mesh drawing function
void MeshInternal::drawMesh(MeshFaceMode faceMode, QOpenGLShaderProgram* prog,
                            const FVector<float,4>& light) {
    // Abort drawing if mesh is empty or uncomplete
    if(vertices.empty())
        return;
    initializeOpenGLFunctions();
    
    prog->setUniformValue("alpha", opacity);
    int vertexLoc = prog->attributeLocation("posAttr");
    int colLoc = prog->attributeLocation("colAttr");
    int normalLoc = prog->attributeLocation("normalAttr");

    glEnableVertexAttribArray(vertexLoc);
    glEnableVertexAttribArray(colLoc);
    glEnableVertexAttribArray(normalLoc);

    if(faceMode == POINT ||
       (!polyline && triangles.empty() && quads.empty())) {
        prog->setUniformValue("light", 0.0, 0.0, 0.0, 0.0);
        glVertexAttribPointer(vertexLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              vertices.data());
        glVertexAttribPointer(colLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              colorsP.data());
        std::vector<FloatVector3> dummy(vertices.size(), FloatVector3(0.0f));
        glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              dummy.data());
        glDrawArrays(GL_POINTS, 0, vertices.size());
    }
    if(polyline && faceMode != POINT) {
        prog->setUniformValue("light", 0.0, 0.0, 0.0, 0.0);
        glVertexAttribPointer(vertexLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              vertices.data());
        glVertexAttribPointer(colLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              colorsP.data());
        std::vector<FloatVector3> dummy(vertices.size(), FloatVector3(0.0f));
        glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              dummy.data());
        glDrawArrays(GL_LINE_STRIP, 0, vertices.size());
    }
    if(faceMode == LINE) {
        prog->setUniformValue("light", 0.0, 0.0, 0.0, 0.0);
        glVertexAttribPointer(vertexLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              verticesTL.data());
        glVertexAttribPointer(colLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              colorsTL.data());
        std::vector<FloatVector3> dummy(std::max(verticesTL.size(), verticesQL.size()),
                                        FloatVector3(0.0f));
        glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              dummy.data());
        glDrawArrays(GL_LINES, 0, verticesTL.size());
        glVertexAttribPointer(vertexLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              verticesQL.data());
        glVertexAttribPointer(colLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              colorsQL.data());
        glDrawArrays(GL_LINES, 0, verticesQL.size());
    }
    if(faceMode == FILL) {
        prog->setUniformValue("light", light[0], light[1], light[2], light[3]);
        glVertexAttribPointer(vertexLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              verticesT.data());
        glVertexAttribPointer(colLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              colorsT.data());
        glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              normalsT.data());
        glDrawArrays(GL_TRIANGLES, 0, verticesT.size());
        glVertexAttribPointer(vertexLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              verticesQ.data());
        glVertexAttribPointer(colLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              colorsQ.data());
        glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0,
                              normalsQ.data());
        glDrawArrays(GL_TRIANGLES, 0, verticesQ.size());
    }

    glDisableVertexAttribArray(vertexLoc);
    glDisableVertexAttribArray(colLoc);
    glDisableVertexAttribArray(normalLoc);
}

// ====================================================================== //
// Utility functions.

// Sum of vertex vectors, useful to compute centroid
FloatPoint3 Mesh::sum(int& nvertices) const {
    nvertices = (int)priv->vertices.size();
    FloatPoint3 s(0,0,0);
    std::vector<FloatPoint3>::const_iterator
        it =priv->vertices.begin(),
        end=priv->vertices.end();
    for(; it!=end; ++it)
        s += *it;
    return s;
}

// Max distance to point
float Mesh::spread(const FloatPoint3& center) const {
    float s=0;
    std::vector<FloatPoint3>::const_iterator
        it =priv->vertices.begin(),
        end=priv->vertices.end();
    for(; it!=end; ++it)
        s = std::max(s, maxNorm(*it-center));
    return s;
}

// ====================================================================== //
// I/O
bool readMeshFromObjFile(Mesh& mesh, const std::string& fileName) {
    // Attempt to read file.
    std::ifstream file(fileName.c_str());
    if(!file) {
        std::cerr << "Error reading OBJ file: " << fileName << std::endl;
        return false;
    }
    
    // Clear the mesh data structure.
    mesh = Mesh();

    // Fill the mesh data structure.
    std::string line;
    while( std::getline(file, line) ) {
        std::stringstream ss;
        ss << line;
        char type;
        ss >> type;
        if(type=='v') {
            float x, y, z;
            ss >> x >> y >> z;
            mesh.priv->vertices.push_back(FloatPoint3(x, y, z));
        }
        if(type=='f') {
            int a, b, c;
            ss >> a >> b >> c;
            mesh.priv->triangles.push_back(Triangle(a-1, b-1, c-1));
        }
    }
    const int n = static_cast<int>(mesh.priv->vertices.size());
    for(std::vector<Triangle>::const_iterator it=mesh.priv->triangles.begin();
        it != mesh.priv->triangles.end(); ++it) {
        int i=0;
        for(; i<3; i++)
            if((*it)[i] < 0 || (*it)[i] >= n)
                break;
        if(i<3) {
            std::cerr << "Warning: wrong vertex index in triangle of OBJ file "
                      << fileName << ": Ignoring all faces" << std::endl;
            mesh.priv->triangles.clear();
            break;
        }
    }   

    mesh.priv->computeNormals();
    mesh.priv->setDataShaders();
    return true;
}

// ======================================================================
// Mesh implementation

Mesh::Mesh()
: priv(new MeshInternal), count(new int(1))
{}

Mesh::Mesh(const FloatPoint3* vertices, int nvertices,
           const Triangle* triangles, int ntriangles,
           const Quad* quads, int nquads,
           MeshColorMode colorMode,
           MeshShadingMode shadingMode,
           const FloatVector3* normals)
: priv(new MeshInternal(vertices,nvertices, triangles,ntriangles, quads,nquads,
                        colorMode, shadingMode, normals))
, count(new int(1))
{}

Mesh::Mesh(const DoublePoint3 *vertices, int nvertices,
           const Triangle* triangles, int ntriangles,
           const Quad *quads, int nquads,
           MeshColorMode colorMode,
           MeshShadingMode shadingMode,
           const DoubleVector3* normals)
: priv(new MeshInternal(vertices,nvertices, triangles,ntriangles, quads,nquads,
                        colorMode, shadingMode, normals))
, count(new int(1))
{}

Mesh::Mesh(const std::vector<FloatPoint3>& vertices,
           const std::vector<Triangle>& triangles,
           const std::vector<Quad>& quads,
           MeshColorMode colorMode,
           MeshShadingMode shadingMode,
           const std::vector<FloatVector3>& normals)
: priv(new MeshInternal(vertices,triangles,quads,colorMode,shadingMode,normals))
, count(new int(1))
{}

Mesh::Mesh(const Mesh& mesh)
: priv(mesh.priv), count(mesh.count)
{
    ++*count;
}

Mesh::~Mesh() {
    kill();
}

void Mesh::kill() {
    if(--*count == 0) {
        delete priv;
        delete count;
    }
}

Mesh& Mesh::operator=(const Mesh& mesh) {
    if(this != &mesh) {
        kill();
        priv = mesh.priv;
        count = mesh.count;
        ++*count;
    }
    return *this;
}

bool Mesh::operator<(const Mesh& mesh) const {
    return (count<mesh.count);
}

void Mesh::drawMesh(MeshFaceMode faceMode, QOpenGLShaderProgram* prog,
                    const FVector<float,4>& light) const {
    priv->drawMesh(faceMode, prog, light);
}

Mesh Mesh::Arrow(const DoublePoint3 &base, const DoubleVector3 &shaft,
                 double r, int resolution) {
    Mesh mesh; delete mesh.priv;
    mesh.priv = MeshInternal::Arrow(base,shaft,r,resolution);
    return mesh;
}

Mesh Mesh::Cone(const DoublePoint3 &base, const DoubleVector3 &axis,
                double r, int resolution) {
    Mesh mesh; delete mesh.priv;
    mesh.priv = MeshInternal::Cone(base,axis,r,resolution);
    return mesh;
}
Mesh Mesh::Cube(const DoublePoint3&center, const DoubleVector3&axis1,
                const DoubleVector3&axis2, const DoubleVector3&axis3) {
    Mesh mesh; delete mesh.priv;
    mesh.priv = MeshInternal::Cube(center,axis1,axis2,axis3);
    return mesh;
}
Mesh Mesh::Cylinder(const DoublePoint3 &center,
                    const DoubleVector3&axis, double r, int res) {
    Mesh mesh; delete mesh.priv;
    mesh.priv = MeshInternal::Cylinder(center,axis,r,res);
    return mesh;
}
Mesh Mesh::Plane(const DoublePoint3&center,
                 const DoubleVector3&axis1, const DoubleVector3&axis2) {
    Mesh mesh; delete mesh.priv;
    mesh.priv = MeshInternal::Plane(center,axis1,axis2);
    return mesh;
}
Mesh Mesh::PointCloud(const DoublePoint3 *points, int n) {
    Mesh mesh; delete mesh.priv;
    mesh.priv = MeshInternal::PointCloud(points,n);
    return mesh;
}
Mesh Mesh::PolyLine(DoublePoint3 *points, int n, bool closed) {
    Mesh mesh; delete mesh.priv;
    mesh.priv = MeshInternal::PolyLine(points,n,closed);
    return mesh;
}
Mesh Mesh::Sphere(const DoublePoint3 &center, double r, int res) {
    Mesh mesh; delete mesh.priv;
    mesh.priv = MeshInternal::Sphere(center,r,res);
    return mesh;
}

} /* namespace Imagine */

#endif
