#include "Imagine/Graphics/ConfigGraphics.h"
#include "OpenGLWindow.h"

#ifdef IMAGINE_OPENGL

#include <QOpenGLContext>
#include <QOpenGLFunctions>

namespace Imagine{
static double FOVY=30.0; // Angle of view in y direction of camera
static float ZOOM_INCREMENT=1.05f; // For mouse wheel and 'A' and 'Z' keys

// ======================================================================
// Shaders
static const char* vertexShader =
    "uniform highp mat4 pvm;\n"
    "uniform highp mat4 vm;\n"
    "uniform lowp float alpha;\n"
    "uniform mediump vec4 light;\n"
    "uniform lowp float pointSize;\n"

    "attribute highp vec4 posAttr;\n"
    "attribute lowp vec3 colAttr;\n"
    "attribute mediump vec3 normalAttr;\n"

    "varying lowp vec4 col;\n"
    "varying mediump vec3 normal;\n"

    "void main() {\n"
    "   col = vec4(colAttr,alpha);\n"
    "   normal = vec3(0.0);\n"
    "   if(light.xyz != vec3(0.0)) {\n"
    "      highp vec4 n = vm * vec4(normalAttr,0.0);\n"
    "      normal = n.xyz; }\n"
    "   gl_Position = pvm * posAttr;\n"
    "   gl_PointSize = pointSize;\n"
    "}\n";

static const char* fragmentShader =
    "uniform mediump vec4 light;\n"
    "varying lowp vec4 col;\n"
    "varying mediump vec3 normal;\n"

    "void main() {\n"
    "   lowp float factor=1.0;\n"
    "   if(light.xyz != vec3(0.0))\n"
    "      factor=abs(dot(light.xyz,normalize(normal)))+light.w;\n"
    "   gl_FragColor = vec4(factor*col.xyz,col.w);\n"
    "}\n";

// ======================================================================
// TrackBall

// Makes vector turn around axis of angle
void rotate(DoubleVector3& vector, DoubleVector3 axis, float angle) {
    double ca = cos(angle);
    double sa = sin(angle);
    DoubleVector3 cross = vector^axis;
    double dot =  axis*vector;
    vector = vector*ca + cross*sa + dot*axis*(1-ca);
}

void TrackBall::move(const QPointF& p, Mode m) {
    DoubleVector3 right = up^eye;
    right.normalize();
    QPointF delta = p-lastPos_;
    if(m==Rotate) {
        rotate(up,right,delta.y());
        rotate(eye,up,delta.x());
        rotate(eye,right,delta.y());
    } else {
        pos += (right*delta.x() - up*delta.y())*norm(eye)/5;
    }
    lastPos_ = p;
}

// ======================================================================
// OpenGLWindow implementation

OpenGLWindow::OpenGLWindow(Imagine::Window win, int subWinId,
                           int width, int height,
                           const QString& windowTitle, int x, int y,
                           QWidget *parent)
: QOpenGLWidget(parent)
, win_(win), subWinId_(subWinId)
, mode_(FILL)
, backgroundColor(156, 156, 255)
, size_(width,height)
, pointSize_(1)
, prog_(0)
{
    float f[4] = {0.0f, 0.0f, 1.0f, 0.2f};
    light_ = FVector<float,4>(f);
    setWindowTitle(windowTitle);
    if(x!=-1 && y!=-1)
        move(x,y);
    resize(width,height);
    show();
}

void OpenGLWindow::capture(QImage* image) {
    *image = grabFramebuffer();
}

void OpenGLWindow::setBackground(const QColor& col) {
    glClearColor(col.red()/255.0f, col.green()/255.0f, col.blue()/255.0f, 1.0f);
    backgroundColor = col;
    update();
}

void OpenGLWindow::centerOnMeshes() {
    FloatPoint3 centerVec(0.0f);
    float radius=0.0f;
    int n=0;
    std::set<Mesh>::iterator it=mesh_.begin(), end=mesh_.end();
    for(; it != end; ++it) {
        int v=0;
        centerVec += it->sum(v);
        n += v;
    }
    if(n) {
        centerVec /= float(n);
        for(it=mesh_.begin(); it != end; ++it) {
            float v = it->spread(centerVec);
            if(radius<v)
                radius=v;
        }
    }
    m_trackBall.pos = centerVec;
    m_trackBall.up = DoubleVector3(0.0,1.0,0.0);
    m_trackBall.eye =
        DoubleVector3(0.0, 0.0, -radius/(float)sin(FOVY/2/180.0*M_PI /1.5));
}

void OpenGLWindow::addMesh(const Mesh& mesh, bool reinitCam) {
    mesh_.insert(mesh);
    if(reinitCam)
        centerOnMeshes();
    update();
}

void OpenGLWindow::removeMesh(const Mesh& mesh, bool reinitCam) {
    mesh_.erase(mesh);
    if(reinitCam)
        centerOnMeshes();
    update();
}

void OpenGLWindow::clear() {
    mesh_.clear();
    update();
}

void OpenGLWindow::setLookAt(double posX, double posY, double posZ,
                             double dirX, double dirY, double dirZ,
                             double upX, double upY, double upZ) {
    m_trackBall.pos = DoubleVector3(posX,posY,posZ);
    m_trackBall.up = DoubleVector3(upX,upY,upZ);
    m_trackBall.eye = DoubleVector3(-dirX,-dirY,-dirZ);
    update();
}

void OpenGLWindow::setLight(float dirx, float diry, float dirz, float ambient) {
    FloatVector3 dir(dirx, diry, dirz);
    if(dir != FloatVector3::Zero())
        dir.normalize();
    float f[4]={dir[0], dir[1], dir[2], ambient};
    light_ = FVector<float,4>(f);
    update();
}

// Drawing tranparent after opaque meshes is necessary, but not sufficient,
// as a transparent mesh drawn behind another transparent mesh will be hidden.
void OpenGLWindow::displayMeshes(const FMatrix<float,4,4>& vm) {
    prog_->bind();
    GLfloat mat[4][4];
    for(int j=0; j<4; j++)
        for(int i=0; i<4; i++)
            mat[i][j] = vm(j,i);
    prog_->setUniformValue("vm", mat);
    FMatrix<float,4,4> pvm = matP_ * vm;
    for(int j=0; j<4; j++)
        for(int i=0; i<4; i++)
            mat[i][j] = pvm(j,i);
    prog_->setUniformValue("pvm", mat);
    prog_->setUniformValue("pointSize", (float)pointSize_);

    std::set<Mesh>::iterator it=mesh_.begin(), end=mesh_.end();
    for(; it != end; ++it) // Draw opaque meshes
        if(it->getOpacity() == 1.0f)
            it->drawMesh(mode_, prog_, light_);
    for(it=mesh_.begin(); it!=end; ++it) // Draw transparent meshes
        if(it->getOpacity() != 1.0f)
            it->drawMesh(mode_, prog_, light_);
    prog_->release();
}

void OpenGLWindow::initializeGL() {
    if(! prog_) {
        prog_ = new QOpenGLShaderProgram(this);
        prog_->addShaderFromSourceCode(QOpenGLShader::Vertex, vertexShader);
        prog_->addShaderFromSourceCode(QOpenGLShader::Fragment, fragmentShader);
        prog_->link();
    }

    QOpenGLContext::currentContext()->functions()->glClearColor(backgroundColor.red()/255.0f, backgroundColor.green()/255.0f, backgroundColor.blue()/255.0f, 1.0f);
    // glShadeModel(GL_SMOOTH); // Enable smooth shading
    // For depth consistent drawing
    QOpenGLContext::currentContext()->functions()->glEnable(GL_DEPTH_TEST);
    // Enable blending (for transparency)
    QOpenGLContext::currentContext()->functions()->glEnable(GL_BLEND);
    QOpenGLContext::currentContext()->functions()->glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // Variable point size
    QOpenGLContext::currentContext()->functions()->glEnable(GL_PROGRAM_POINT_SIZE);
}

void OpenGLWindow::paintGL() {
    QOpenGLContext::currentContext()->functions()->glClearColor(backgroundColor.red()/255.0f, backgroundColor.green()/255.0f, backgroundColor.blue()/255.0f, 1.0f);
    QOpenGLContext::currentContext()->functions()->glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Setting the lookAt
    // gluLookAt: eye, center, up
    // The following is equivalent, whithout requiring glu, to
    // gluLookAt(m_trackBall.pos.x()-m_trackBall.eye.x(),
    //           m_trackBall.pos.y()-m_trackBall.eye.y(),
    //           m_trackBall.pos.z()-m_trackBall.eye.z(),
    //           m_trackBall.pos.x(), m_trackBall.pos.y(), m_trackBall.pos.z(),
    //           m_trackBall.up.x(), m_trackBall.up.y(), m_trackBall.up.z());
    DoubleVector3 f = m_trackBall.eye / norm(m_trackBall.eye);
    DoubleVector3 up = m_trackBall.up / norm(m_trackBall.up);
    DoubleVector3 s=f^up, u=s^f;
    GLfloat M[4*4] = { // Transposed, as OpenGL expects column major matrices
        (GLfloat)s.x(), (GLfloat)u.x(), -(GLfloat)f.x(), 0.0f,
        (GLfloat)s.y(), (GLfloat)u.y(), -(GLfloat)f.y(), 0.0f,
        (GLfloat)s.z(), (GLfloat)u.z(), -(GLfloat)f.z(), 0.0f,
                  0.0f,           0.0f,            0.0f, 1.0f};

    FMatrix<float,4,4> vm = FMatrix<float,4,4>::Identity();
    vm(0,3) = m_trackBall.eye.x()-m_trackBall.pos.x();
    vm(1,3) = m_trackBall.eye.y()-m_trackBall.pos.y();
    vm(2,3) = m_trackBall.eye.z()-m_trackBall.pos.z();
    vm = FMatrix<float,4,4>(M) * vm; // View-Model matrix
    displayMeshes(vm);
}

void OpenGLWindow::resizeGL(int width, int height) {
    QOpenGLContext::currentContext()->functions()->glViewport(0, 0, width, height);

    // The following is equivalent, without requiring glu, to
    // gluPerspective(FOVY, ratio, 0.1, 1000.0);
    // An alternative is also:
    // glFrustum(-fW,fW, -fH,fH, ZNear, ZFar);
    const GLfloat ZNear=0.1f;
    const GLfloat ZFar=1000.0f;
    double ratio = width/static_cast<double>(height);
    GLfloat fH = tan(FOVY/2 /180.0*M_PI)*ZNear, fW=fH*ratio;

    // See http://www.songho.ca/opengl/gl_projectionmatrix.html
    GLfloat M[4*4] = { // transposed
        2*ZNear / (2*fW), 0, 0, 0,
        0, 2*ZNear / (2*fH), 0, 0,
        0, 0, -(ZFar+ZNear)/(ZFar-ZNear), -1,
        0, 0, -2*ZFar*ZNear/(ZFar-ZNear), 0};
    matP_ = FMatrix<float,4,4>(M);
}

QPointF OpenGLWindow::pixelPosToViewPos(const QPointF& p) {
    return QPointF(2.0 * float(p.x()) / width() - 1.0,
                   1.0 - 2.0 * float(p.y()) / height());
}

void OpenGLWindow::mousePressEvent(QMouseEvent *event) {
    QOpenGLWidget::mousePressEvent(event);
    if (event->isAccepted())
        return;
    if(event->modifiers()==Qt::ShiftModifier) {
        if(event->button()==Qt::LeftButton ||
           event->button()==Qt::RightButton) {
#if QT_VERSION < 0x050000
            m_trackBall.push(pixelPosToViewPos(event->posF()));
#else
            m_trackBall.push(pixelPosToViewPos(event->localPos()));
#endif
            event->accept();
        }
    }
    if(event->isAccepted())
        update();
    else
        emit pressedMouseButtons(event->x(), event->y(), event->button(),
                                 event->type(), event->modifiers(),
                                 win_, subWinId_);
}

void OpenGLWindow::mouseReleaseEvent(QMouseEvent *event) {
    QOpenGLWidget::mouseReleaseEvent(event);
    if(! event->isAccepted())
        emit pressedMouseButtons(event->x(), event->y(), event->button(),
                                 event->type(), event->modifiers(),
                                 win_, subWinId_);
}

void OpenGLWindow::mouseMoveEvent(QMouseEvent *event) {
    QOpenGLWidget::mouseMoveEvent(event);
    if (event->isAccepted())
        return;

    if(event->modifiers()==Qt::ShiftModifier) {
        if(event->buttons() & Qt::LeftButton) {
#if QT_VERSION < 0x050000
            m_trackBall.move(pixelPosToViewPos(event->posF()),
                             TrackBall::Rotate);
#else
            m_trackBall.move(pixelPosToViewPos(event->localPos()),
                             TrackBall::Rotate);
#endif
            event->accept();
        }
        if(event->buttons() & Qt::RightButton) {
#if QT_VERSION < 0x050000
            m_trackBall.move(pixelPosToViewPos(event->posF()),
                             TrackBall::Translate);
#else
            m_trackBall.move(pixelPosToViewPos(event->localPos()),
                             TrackBall::Translate);
#endif
            event->accept();
        }
    }
    if(event->isAccepted())
        update();
    else
        emit pressedMouseButtons(event->x(), event->y(), event->button(),
                                 event->type(), event->modifiers(),
                                 win_, subWinId_);
}

void OpenGLWindow::wheelEvent(QWheelEvent* event) {
    QOpenGLWidget::wheelEvent(event);
    if(!event->isAccepted()) {
        if(event->angleDelta().y()>0)
            m_trackBall.eye *= ZOOM_INCREMENT;
        else
            m_trackBall.eye /= ZOOM_INCREMENT;
        event->accept();
        update();
    }
}

void OpenGLWindow::keyPressEvent(QKeyEvent* event) {
    if(event->modifiers()==Qt::ShiftModifier &&
       event->key() == 'M') { // Switch display mode
        if(mode_ == POINT) mode_=FILL;
        else if(mode_ == LINE) mode_=POINT;
        else if(mode_ == FILL) mode_=LINE;
        event->accept();
        update();
    } else if(event->modifiers()==Qt::ShiftModifier &&
              event->key() == 'A') { // Zoom in
        m_trackBall.eye *= ZOOM_INCREMENT;
        event->accept();
        update();
    } else if(event->modifiers()==Qt::ShiftModifier &&
              event->key() == 'Z') { // Zoom out
        m_trackBall.eye /= ZOOM_INCREMENT;
        event->accept();
        update();
    } else if(event->modifiers()==Qt::ShiftModifier &&
              event->key() == 'O') { // Decrease point size
        event->accept();
        if(pointSize_ > 1) {
            --pointSize_;
            update();
        }
    } else if(event->modifiers()==Qt::ShiftModifier &&
              event->key() == 'P') { // Increase point size
        event->accept();
        ++pointSize_;
        update();
    } else if(event->modifiers()==Qt::ShiftModifier &&
              event->key() == 'R') { // Reinit camera
       centerOnMeshes();
       event->accept();
       update();
    }else
        emit pressedKey(event->key(), event->type(), event->modifiers(),
                        win_, subWinId_);
}

void OpenGLWindow::keyReleaseEvent(QKeyEvent* event) {
    if(event->isAutoRepeat())
        QWidget::keyReleaseEvent(event);
    else
        emit pressedKey(event->key(), event->type(), event->modifiers(),
                        win_, subWinId_);
}

void OpenGLWindow::closeEvent(QCloseEvent* event) {
    if(event->spontaneous())
        exit(0);
    emit closed();
    this->QWidget::closeEvent(event);
}

} /* namespace Imagine */

#endif
